from .base import *
from ..qf_map.u_lyr_map import *

class U_LYR_Circ(LinnerCircuit):
    """
    U_LYR_Circ is a class, which includes functions to build u-layer

    Args:
         n_qubits: input qubits of each unit
         n_repeats: repeat times of each unit

    """

    def __init__(self, n_qubits, n_repeats):
        LinnerCircuit.__init__(self, n_qubits, n_repeats)

    @classmethod
    def extract_from_weight(self, weights):
        """
        Function extract_from_weight : transform the weights of the nets into the gates of qubits.

        Args:
             weights: the weights of nets
        Returns:
             gates: The mapped gate list ,such as 0010.

        """
        # Find Z control gates according to weights
        w = (weights.detach().cpu().numpy())
        total_len = len(w)
        target_num = np.count_nonzero(w == -1)
        if target_num > total_len / 2:
            w = w * -1
        target_num = np.count_nonzero(w == -1)
        digits = int(math.log(total_len, 2))
        flag = "0" + str(digits) + "b"
        max_num = int(math.pow(2, digits))
        sign = {}
        for i in range(max_num):
            sign[format(i, flag)] = +1

        quantum_gates = Mapping_U_LYR(sign, target_num, digits)

        # Build the mapping from weight to final negative num
        fin_sign = list(sign.values())
        fin_weig = [int(x) for x in list(w)]
        sign_neg_index = []
        try:
            beg_pos = 0
            while True:
                find_pos = fin_sign.index(-1, beg_pos)
                # qiskit_position = int(format(find_pos,flag)[::-1],2)
                sign_neg_index.append(find_pos)
                beg_pos = find_pos + 1
        except Exception as exception:
            pass

        weight_neg_index = []
        try:
            beg_pos = 0
            while True:
                find_pos = fin_weig.index(-1, beg_pos)
                weight_neg_index.append(find_pos)
                beg_pos = find_pos + 1
        except Exception as exception:
            pass

        map = {}
        for i in range(len(sign_neg_index)):
            map[sign_neg_index[i]] = weight_neg_index[i]

        ret_index = list([-1 for i in range(len(fin_weig))])

        for k, v in map.items():
            ret_index[k] = v

        for i in range(len(fin_weig)):
            if ret_index[i] != -1:
                continue
            for j in range(len(fin_weig)):
                if j not in ret_index:
                    ret_index[i] = j
                    break
        return quantum_gates, ret_index


# if __name__ == "__main__":
#     import torch
#     from qiskit import QuantumRegister, QuantumCircuit
#     weight_1 = torch.tensor([[1., 1., 1., 1., 1., 1., 1., -1., 1., -1., 1., -1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., -1., 1., -1., 1., -1., 1., 1., 1., 1.],
#                              [-1., -1., -1., -1., -1., -1., -1., -1., -1., 1., -1., 1., -1., -1., -1., -1.,-1., -1., -1., -1., -1., -1., -1., -1., -1., 1., -1., 1., -1., -1., -1., -1.]])
#     #define your input and output number
#     u_layer = U_LYR_Circ(5,2)
#     #init circuit
#     opt_circ = QuantumCircuit()
#     #add input qubit to your circuit
#     inps = u_layer.add_input_qubits(opt_circ)
#     #add auxiliary qubit to your circuit
#     aux =u_layer.add_aux(opt_circ)
#     #add output qubit to your circuit
#     u_layer_out_qubits = u_layer.add_out_qubits(opt_circ)
#     quantum_matrix = torch.eye(32)
#     #add ulayer to your circuit
#     u_layer.forward(opt_circ,weight_1,inps,u_layer_out_qubits,quantum_matrix,aux)
#     print(opt_circ)


from .base import *

class F_LYR_Circ(LinnerCircuit):
    """
    F_LYR_Circ is a class, which includes functions to build F-layer(the former part of FFNN)

    Args:
         n_qubits: input qubits of each unit
         n_repeats: repeat times of each unit

    """
    def __init__(self, n_qubits, n_repeats):

        LinnerCircuit.__init__(self, n_qubits, n_repeats)

    @classmethod
    def AinB(self, A, B):
        idx_a = []
        for i in range(len(A)):
            if A[i] == "1":
                idx_a.append(i)
        flag = True
        for j in idx_a:
            if B[j] == "0":
                flag = False
                break
        return flag

    @classmethod
    def extract_from_weight(self, weights):
        """
        Function extract_from_weight : transform the weights of the nets into the gates of qubits.

        Args:
             weights: the weights of nets
        Returns:
             gates: The mapped gate list ,such as 0010.

        """
        # Find Z control gates according to weights
        w = (weights.detach().cpu().numpy())
        total_len = len(w)
        digits = int(math.log(total_len, 2))
        flag = "0" + str(digits) + "b"
        max_num = int(math.pow(2, digits))
        sign = {}
        for i in range(max_num):
            sign[format(i, flag)] = +1
        sign_expect = {}
        for i in range(max_num):
            sign_expect[format(i, flag)] = int(w[i])

        order_list = []
        for i in range(digits + 1):
            for key in sign.keys():
                if key.count("1") == i:
                    order_list.append(key)

        gates = []
        sign_cur = copy.deepcopy(sign_expect)
        for idx in range(len(order_list)):
            key = order_list[idx]
            if sign_cur[key] == -1:
                gates.append(key)
                for cor_idx in range(idx, len((order_list))):
                    if self.AinB(key, order_list[cor_idx]):
                        sign_cur[order_list[cor_idx]] = (-1) * sign_cur[order_list[cor_idx]]
        return gates, None





from .utils import *

#p-layer
class P_LYR(nn.Linear):

    def do_slp_via_th(self, input_ori, w_ori):
        p = input_ori
        d = 4 * p * (1 - p)
        e = (2 * p - 1)
        # e_sq = torch.tensor(1)
        w = w_ori

        sum_of_sq = (d + e.pow(2)).sum(-1)
        sum_of_sq = sum_of_sq.unsqueeze(-1)
        sum_of_sq = sum_of_sq.expand(p.shape[0], w.shape[0])

        diag_p = torch.diag_embed(e)

        p_w = torch.matmul(w, diag_p)

        z_p_w = torch.zeros_like(p_w)
        shft_p_w = torch.cat((p_w, z_p_w), -1)

        sum_of_cross = torch.zeros_like(p_w)
        length = p.shape[1]

        for shft in range(1, length):
            sum_of_cross += shft_p_w[:, :, 0:length] * shft_p_w[:, :, shft:length + shft]

        sum_of_cross = sum_of_cross.sum(-1)

        return (sum_of_sq + 2 * sum_of_cross) / (length ** 2)

    def forward(self, input, training=True):
        binary_weight = binarize(self.weight)
        if self.bias is None:
            # print(input,binary_weight,self.do_slp_via_th(input, binary_weight))
            # sys.exit(0)
            return self.do_slp_via_th(input, binary_weight)

        else:

            bias_one = torch.ones(input.shape[0], 1)
            new_input = torch.cat((input, bias_one), -1)
            bias = clipfunc(self.bias).unsqueeze(1)
            new_weight = binary_weight
            new_weight = torch.cat((new_weight, bias), -1)
            return self.do_slp_via_th(new_input, new_weight)


    def reset_parameters(self):
        # Glorot initialization
        in_features, out_features = self.weight.size()
        stdv = math.sqrt(1.5 / (in_features + out_features))
        self.weight.data.uniform_(-stdv, stdv)
        if self.bias is not None:
            self.bias.data.zero_()

        self.weight.lr_scale = 1. / stdv




import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Function
import numpy as np
import math
import sys
from torch.nn.parameter import Parameter
import torch
import os
import sys
import shutil


class BinarizeF(Function):
    '''
    >0: 1 <0:-1
    '''

    @staticmethod
    def forward(cxt, input):
        
        output = input.new(input.size())
        output[input >= 0] = 1
        output[input < 0] = -1

        return output

    @staticmethod
    def backward(cxt, grad_output):
        grad_input = grad_output.clone()
        return grad_input

class ClipF(Function):

    @staticmethod
    def forward(ctx, input):
        output = input.clone().detach()
        # output = input.new(input.size())
        output[input >= 1] = 1
        output[input <= 0] = 0
        ctx.save_for_backward(input)
        return output

    @staticmethod
    def backward(ctx, grad_output):
        input, = ctx.saved_tensors
        grad_input = grad_output.clone()
        grad_input[input >= 1] = 0
        grad_input[input <= 0] = 0
        return grad_input


def save_checkpoint(state, is_best, save_path, filename):
    filename = os.path.join(save_path, filename)
    torch.save(state, filename)
    if is_best:
        bestname = os.path.join(save_path, 'model_best.tar')
        shutil.copyfile(filename, bestname)


def qf_sum(n_qubits):
    sum_mat = []
    flag = "0" + str(n_qubits) + "b"
    for i in range(0, int(math.pow(2, n_qubits))):
        bit_str = format(i, flag)
        row = []
        for c in bit_str:
            row.append(float(c))
        sum_mat.append(row)
    return sum_mat


def set_value(mm, col, row, val):
    index = (torch.LongTensor([col]), torch.LongTensor([row]))  # 生成索引
    mm = mm.index_put(index, val)
    return mm


def tensor_squire(x):
    return torch.pow(x, 2)


def tensor_sqrt(x):
    return torch.sqrt(x + 1e-8)


def amp2prop(state):
    state = state.double()
    state = state * state
    n_qubits = int(math.log2(state.shape[0]))
    sum_mat = torch.tensor(qf_sum(n_qubits), dtype=torch.float64)
    sum_mat = sum_mat.t()
    state = torch.mm(sum_mat, state)
    return state


class Prob2amp():
    def __call__(self, state,training=False):
        state = state.double().t()
        n_qubits = state.shape[0]
        mstate = torch.ones(int(math.pow(2, n_qubits)), state.shape[1], dtype=torch.float64)
        sum_mat = torch.tensor(qf_sum(n_qubits), dtype=torch.float64)
        for i in range(sum_mat.shape[0]):
            for j in range(sum_mat.shape[1]):
                if int(sum_mat[i][j]) == 0:
                    val = torch.mul(torch.index_select(mstate, 0, torch.tensor([i])),
                                    1 - torch.index_select(state, 0, torch.tensor([n_qubits-1-j]))).squeeze()
                    mstate = set_value(mstate, i, range(state.shape[1]), val)
                elif int(sum_mat[i][j]) == 1:
                    val = torch.mul(torch.index_select(mstate, 0, torch.tensor([i])),
                                    torch.index_select(state, 0, torch.tensor([n_qubits-1-j]))).squeeze()
                    mstate = set_value(mstate, i, range(state.shape[1]), val)

        # print(mstate.t())
        mstate = tensor_sqrt(mstate)
        return mstate.t()


# instances
binarize = BinarizeF.apply
clipfunc = ClipF.apply


from ..qf_net.n_lyr import *
from ..qf_net.p_lyr import *
from ..qf_net.u_lyr import *
from ..qf_net.v_lyr import *
from ..qf_net.b_lyr import *

## Define the NN architecture
class Net(nn.Module):
    """
    class Net is to simply build up a network consisting of quantum layers ,using mat to represent the quantum gate.

    Args:
         img_size: the width /height of input image (width = height)
         layers: a 2-dimensions list. for example,[['u',4]['p',2]] means that the first layer is u-layer with 4 output qubit, and the second layer is p-layer with 2 output qubit
         training: whether  training
         binary: whether the input data should be binarized
         given_ang: initial angle for N-layer if used

    """
    def __init__(self,img_size,layers,training,binary,given_ang=[],train_ang=False,debug="False"):
        super(Net, self).__init__()

        self.in_size = img_size*img_size
        self.training = training
        self.layer = len(layers)
        self.layers = layers
        self.binary = binary
        loop_in_size = self.in_size
        self.debug = debug
        for idx in range(self.layer):
            fc_name = "fc"+str(idx)
            if layers[idx][0]=='u':
                setattr(self, fc_name, U_LYR(loop_in_size, layers[idx][1], bias=False))
            elif layers[idx][0]=='c':
                setattr(self, fc_name, B_LYR(loop_in_size, layers[idx][1], bias=False))
            elif layers[idx][0]=='f':
                setattr(self, fc_name, nn.Linear(loop_in_size, layers[idx][1]))
            elif layers[idx][0]=='p':
                setattr(self, fc_name, P_LYR(loop_in_size, layers[idx][1], bias=False))
            elif layers[idx][0]=='p2a':
                setattr(self, fc_name, Prob2amp())
            elif layers[idx][0]=='v':
                setattr(self, fc_name, VQC_Net(loop_in_size, layers[idx][1]))
            elif layers[idx][0]=='v10':
                setattr(self, fc_name, VQC_Net(loop_in_size, layers[idx][1],'vqc_10'))
            elif layers[idx][0]=='v5':
                setattr(self, fc_name, VQC_Net(loop_in_size, layers[idx][1],'vqc_5'))
            elif layers[idx][0]=='n':
                setattr(self, fc_name, N_LYR(num_features=layers[idx][1], init_ang_inc=given_ang[idx], training=train_ang))
            else:
                print("Not support layer name!")
                sys.exit(0)
            loop_in_size = layers[idx][1]


    def forward(self, x, training=1):
        x = x.view(-1, self.in_size)
        for layer_idx in range(self.layer):
            if self.binary and layer_idx==0:
                x = (binarize(x - 0.5) + 1) / 2
            x = getattr(self, "fc" + str(layer_idx))(x,training)

        if self.layers[-1][1] == 1:
            x = torch.cat((x, 1 - x), -1)

        return x





from qiskit.tools.monitor import job_monitor
from qiskit import QuantumRegister,ClassicalRegister
from qiskit import Aer, execute,IBMQ,transpile
from qiskit import BasicAer

def auth_output():
    print("welcome to Weiwen Jiang's Quantumflow")


################ Weiwen on 12-30-2020 ################
# Function: fire_ibmq from Listing 6
# Note: used for execute quantum circuit using
#       simulation or ibm quantum processor
# Parameters: (1) quantum circuit;
#             (2) number of shots;
#             (3) simulation or quantum processor;
#             (4) backend name if quantum processor.
######################################################
def fire_ibmq(circuit, shots, Simulation=True, backend_name='ibmq_essex'):
    """
    class fire_ibmq is used for execute quantum circuit using simulation or ibm quantum processor

    Args:
         circuit: quantum circuit
         shots: number of shots
         Simulation: simulation or quantum processor
         backend_name: backend name if quantum processor

    Returns:
         counts: quantum execute results, including different states with shots
    """

    if not Simulation:
        provider = IBMQ.get_provider('ibm-q-academic')
        backend = provider.get_backend(backend_name)
    else:
        backend = Aer.get_backend('qasm_simulator')
    # circuit.save_statevector()
    job_ibm_q = execute(circuit, backend, shots=shots)
    if not Simulation:
        job_monitor(job_ibm_q)
    result_ibm_q = job_ibm_q.result()

    counts = result_ibm_q.get_counts()

    return counts


def get_unitary(circuit, IBMQ=None):
    """
    Function get_unitary is used for execute quantum circuit using simulation to get the unitary matrix that represents the circuit

    Args:
         circuit: quantum circuit

    Returns:
         unitary: unitary matrix
    """
    if IBMQ == None:
        backend = BasicAer.get_backend('unitary_simulator')
    else:
        provider = IBMQ.get_provider(hub='ibm-q', group='open', project='main')
        backend = provider.backend.ibmq_vigo

    job = backend.run(transpile(circuit, backend))

    unitary = job.result().get_unitary(circuit, decimals=9)  # Execute the circuit
    return unitary


def add_measure(circuit, qubits, name):
    """
    Function add_measure is used for add the measure gate to the qubits

    Args:
         circuit: quantum circuit
         qubits: input quantum registers
    """
    lenth = len(qubits)
    c_reg = ClassicalRegister(lenth, name)
    circuit.add_register(c_reg)
    for i in range(lenth):
        circuit.measure(qubits[i], c_reg[i])


################ Weiwen on 12-30-2020 ################
# Function: analyze from Listing 6
# Note: used for analyze the count on states to
#       formulate the probability for each qubit
# Parameters: (1) counts returned by fire_ibmq;
######################################################
def analyze(counts):
    """
    Function analyze is used for analyze the count on states to
       formulate the probability for each qubit
    
    Args:
         counts: : quantum execute results, including different states with shots

    Returns:
         mycount:  shots of the state '1' for each qubit
         bits: the number of qubits
    """
    mycount = {}
    for i in range(2):
        mycount[i] = 0
    for k, v in counts.items():
        bits = len(k)
        for i in range(bits):
            if k[bits - 1 - i] == "1":
                if i in mycount.keys():
                    mycount[i] += v
                else:
                    mycount[i] = v
    return mycount, bits


# -*- encoding: utf-8 -*-
'''
Filename         :lib_circuit_base.py
Description      :This document is used for fundamental class of quantum circuit
Time             :2021/09/26 13:58:23
Author           :Weiwen Jiang & Zhirui Hu
Version          :1.0
'''


import sys
import abc
from qiskit import QuantumRegister
from qiskit.extensions import UnitaryGate
from .gates import *
import copy
import numpy as np
import math

class BaseCircuit(metaclass= abc.ABCMeta):
    """
    BaseCircuit is a class, which includes fundamental functions of a circuit module.

    Args:
         n_qubits: input qubits of each unit
         n_repeats: repeat times of each unit

    """
    def __init__(self,n_qubits,n_repeats):
        self.n_qubits = n_qubits
        self.n_repeats = n_repeats

    def add_qubits(self,circuit,name,number):
        """
        Function add_qubits is to add a group of qubits to a circuit. [Test at 09/29]

        Args:
            circuit: The circuit that you add the unit at the end
            name: The name of the group
            number:  The number of qubits in the group.

        Returns:
            qubits: The register of qubits

        """
        qubits = QuantumRegister(number,name)
        circuit.add_register(qubits)
        return qubits


    def add_input_qubits(self,circuit,name):
        """
        Function add_input_qubits is to add a group of qubits as input qubit .

        Args:
             circuit: The  circuit that you add the unit at the end
             name: The name of the group

        Returns:
             qubits: The register of qubits

        """
        inps = []
        for i in range(self.n_repeats):
            inp = QuantumRegister(self.n_qubits,name+str(i)+"_qbit")
            circuit.add_register(inp)
            inps.append(inp)
        return inps


    @abc.abstractclassmethod
    def forward(self,circuit):
        """
        Function forward is to add a circuit unit to the input qubits.

        Args:
             circuit: The  circuit that you add the unit at the end
        """
        pass


class LinnerCircuit(BaseCircuit):
    """
    LinnerCircuit is a base class of u-layer circuit and ffnn circuit.

    Args:
        n_qubits: input qubits of each unit
        n_repeats: repeat times of each unit
    """
    def __init__(self, n_qubits, n_repeats):
        self.n_qubits = n_qubits
        self.n_repeats = n_repeats
        # if self.n_qubits > 4:
        #     print('The input size is too big. Qubits should be less than 4.')
        #     sys.exit(0)

    def add_aux(self, circuit):
        """
        Function add_aux  is to add a group of qubits as input qubit .

        Args:
             circuit: The  circuit that you add the unit at the end
        Returns:
             qubits: The register of qubits

        """
        if self.n_qubits < 3:
            aux = self.add_qubits(circuit, "aux_qbit", 1)
        # TODO: 09/30, Potential bug.
        elif self.n_qubits >= 3:
            aux = self.add_qubits(circuit, "aux_qbit", self.n_qubits-2)
        # else:
        #     print('The input size is too big. Qubits should be less than 4.')
        #     sys.exit(0)
        return aux

    def add_input_qubits(self, circuit,name = "linner_layer"):
        """
        Function add_input_qubits is to add a group of qubits as input qubit .

        Args:
             circuit: The  circuit that you add the unit at the end
        Returns:
             qubits: The register of qubits

        """
        inputs = BaseCircuit.add_input_qubits(self, circuit, name)
        return inputs

    def add_out_qubits(self, circuit):
        """
        Function add_out_qubits is to add a group of qubits as input qubit .

        Args:
             circuit: The  circuit that you add the unit at the end
        Returns:
             qubits: The register of qubits

        """
        out_qubits = self.add_qubits(circuit, "u_layer_output_qbit", self.n_repeats)
        return out_qubits

    #@abc.abstractclassmethod
    def extract_from_weight(weight):
        pass

    def add_weight(self, circuit, weight, in_qubits, data_matrix=None, aux=[]):
        """
        Function add_weight is to add the circuit of ci=wi*xi.

        Args:
             circuit: The  circuit that you add the unit at the end
             weight: A list of binary weight.
             in_qubits: The register of input qubits
             data_matrix:The unitary matrix encoding into the qubits
             aux: aux qubits
        """
        for i in range(self.n_repeats):
            n_q_gates, n_idx = self.extract_from_weight(weight[i])
            if data_matrix != None and n_idx != None:
                circuit.append(UnitaryGate(data_matrix[n_idx], label="Input"), in_qubits[i][0:self.n_qubits])
            qbits = in_qubits[i]
            for gate in n_q_gates:
                z_count = gate.count("1")
                # z_pos = get_index_list(gate,"1")
                z_pos = self.get_index_list(gate[::-1], "1")
                # print("z_count= ",z_count)
                if z_count == 0:
                    continue
                elif z_count == 1:
                    circuit.z(qbits[z_pos[0]])
                elif z_count == 2:
                    circuit.cz(qbits[z_pos[0]], qbits[z_pos[1]])
                else:
                    operate_qubits = []
                    aux_qubits = []
                    for k in range(z_count):
                        operate_qubits.append(qbits[z_pos[k]])
                        if k < z_count - 2:
                            aux_qubits.append(aux[k])
                    ExtendGate.cnz(circuit, operate_qubits, aux_qubits, z_count)
                # elif z_count == 3:
                #     ExtendGate.ccz(circuit, qbits[z_pos[0]], qbits[z_pos[1]], qbits[z_pos[2]], aux[0])
                # elif z_count == 4:
                #     ExtendGate.cccz(circuit, qbits[z_pos[0]], qbits[z_pos[1]], qbits[z_pos[2]], qbits[z_pos[3]], aux[0],
                #                     aux[1])
                # else:
                #     print("Not support yet!")
                #     sys.exit(0)
        circuit.barrier()

    def sum2(self, circuit, in_qubits, out_qubit, aux=[]):
        """
        Function sum2 is to add the circuit of (∑ci)^2/N.

        Args:
             circuit: The  circuit that you add the unit at the end
             in_qubits: The register of input qubits
             out_qubit: The register of output qubits
             aux: aux qubits
        """
        for i in range(self.n_repeats):
            circuit.h(in_qubits[i])
            circuit.x(in_qubits[i])
        circuit.barrier()
        for i in range(self.n_repeats):
            qbits = in_qubits[i]
            if self.n_qubits == 1:
                circuit.cx(qbits[0], out_qubit[i])
            elif self.n_qubits == 2:
                circuit.ccx(qbits[0], qbits[1], out_qubit[i])
            else:
                operate_qubits = []
                aux_qubits = []
                for k in range(self.n_qubits):
                    operate_qubits.append(qbits[k])
                    if k < self.n_qubits - 2:
                        aux_qubits.append(aux[k])
                operate_qubits.append(out_qubit[i])
                ExtendGate.cnx(circuit, operate_qubits, aux_qubits, self.n_qubits+1)
            #
            # elif self.n_qubits == 3:
            #     ExtendGate.cccx(circuit, qbits[0], qbits[1], qbits[2], out_qubit[i], aux[0], aux[1])
            # elif self.n_qubits == 4:
            #     ExtendGate.ccccx(circuit, qbits[0], qbits[1], qbits[2], qbits[3], out_qubit[i], aux[0], aux[1])

    def forward(self, circuit, weight, in_qubits, out_qubit, data_matrix=None, aux=[]):
        """
        Function forward is to add the circuit of (∑wi*xi)^2/N.

        Args:
             circuit: The  circuit that you add the unit at the end
             weight: A list of binary weight.
             in_qubits: The register of input qubits
             out_qubit: The register of output qubits
             data_matrix:The unitary matrix encoding into the qubits
             aux: aux qubits
        """
        self.add_weight(circuit, weight, in_qubits, data_matrix, aux)
        self.sum2(circuit, in_qubits, out_qubit, aux)

    @classmethod
    def get_index_list(self, input, target):
        index_list = []
        try:
            beg_pos = 0
            while True:
                find_pos = input.index(target, beg_pos)
                index_list.append(find_pos)
                beg_pos = find_pos + 1
        except Exception as exception:
            pass
        return index_list




from ..qf_circ.base import *
from qiskit.extensions import UnitaryGate

class UMatrixCircuit(BaseCircuit):
    """
    UMatrixCircuit is a class, which encodes unitary matrix into qubits using UnitaryGate

    Args:
         n_qubits: input qubits of each unit
         n_repeats: repeat times of each unit

    """
    def __init__(self,n_qubits,n_repeats):
        self.n_qubits = n_qubits
        self.n_repeats = n_repeats

    def add_input_qubits(self,circuit):
        inps = BaseCircuit.add_input_qubits(self,circuit,"u_matrix_input")
        return inps

    def forward(self,circuit,input_qubits,data_matrix,ids = None):
        for i in range(self.n_repeats):
            if ids == None:
                circuit.append(UnitaryGate(data_matrix, label="Input"), input_qubits[i][0:self.n_qubits])
            else:
                circuit.append(UnitaryGate(data_matrix[ids[i]], label="Input"), input_qubits[i][0:self.n_qubits])



from __future__ import print_function
from torchvision.datasets import VisionDataset
import warnings
from PIL import Image
import os
import os.path
import numpy as np
import torch
import codecs



class qfMNIST(VisionDataset):


    classes = ['0 - zero', '1 - one', '2 - two', '3 - three', '4 - four',
               '5 - five', '6 - six', '7 - seven', '8 - eight', '9 - nine']

    @property
    def train_labels(self):
        warnings.warn("train_labels has been renamed targets")
        return self.targets

    @property
    def test_labels(self):
        warnings.warn("test_labels has been renamed targets")
        return self.targets

    @property
    def train_data(self):
        warnings.warn("train_data has been renamed data")
        return self.data

    @property
    def test_data(self):
        warnings.warn("test_data has been renamed data")
        return self.data

    def __init__(self, root, train=True, img_size=4):
        super(qfMNIST, self).__init__(root)
        self.train = train  # training set or test set

        train_path = root + "/qc_training_" + str(img_size) + "_" + str(img_size) + ".pt"
        test_path = root + "/qc_test_" + str(img_size) + "_" + str(img_size) + ".pt"
        print(train_path,os.path.isfile(train_path))
        print(test_path,os.path.isfile(test_path))
        if not (os.path.isfile(train_path) and os.path.isfile(test_path)):
            raise RuntimeError('Dataset not found.' +
                               ' You can use download=True to download it')

        if self.train:
            data_file = train_path
        else:
            data_file = test_path

        self.data, self.targets = torch.load(data_file)

    def __getitem__(self, index):
        """
        Args:
            index (int): Index

        Returns:
            tuple: (image, target) where target is index of the target class.
        """
        img, target = self.data[index], int(self.targets[index])

        # doing this so that it is consistent with all other datasets
        # to return a PIL Image
        # img = Image.fromarray(img.numpy(), mode='L')

        # if self.transform is not None:
        #     img = self.transform(img)
        #
        # if self.target_transform is not None:
        #     target = self.target_transform(target)

        return img, target

    def __len__(self):
        return len(self.data)

    @property
    def raw_folder(self):
        return os.path.join(self.root, self.__class__.__name__, 'raw')

    @property
    def processed_folder(self):
        return os.path.join(self.root, self.__class__.__name__, 'processed')

    @property
    def class_to_idx(self):
        return {_class: i for i, _class in enumerate(self.classes)}

    def _check_exists(self):
        return (os.path.exists(os.path.join(self.processed_folder,
                                            self.training_file)) and
                os.path.exists(os.path.join(self.processed_folder,
                                            self.test_file)))

    def extra_repr(self):
        return "Split: {}".format("Train" if self.train is True else "Test")



def get_int(b):
    return int(codecs.encode(b, 'hex'), 16)


def open_maybe_compressed_file(path):
    """Return a file object that possibly decompresses 'path' on the fly.
       Decompression occurs when argument `path` is a string and ends with '.gz' or '.xz'.
    """
    if not isinstance(path, torch._six.string_classes):
        return path
    if path.endswith('.gz'):
        import gzip
        return gzip.open(path, 'rb')
    if path.endswith('.xz'):
        import lzma
        return lzma.open(path, 'rb')
    return open(path, 'rb')


def read_sn3_pascalvincent_tensor(path, strict=True):
    """Read a SN3 file in "Pascal Vincent" format (Lush file 'libidx/idx-io.lsh').
       Argument may be a filename, compressed filename, or file object.
    """
    # typemap
    if not hasattr(read_sn3_pascalvincent_tensor, 'typemap'):
        read_sn3_pascalvincent_tensor.typemap = {
            8: (torch.uint8, np.uint8, np.uint8),
            9: (torch.int8, np.int8, np.int8),
            11: (torch.int16, np.dtype('>i2'), 'i2'),
            12: (torch.int32, np.dtype('>i4'), 'i4'),
            13: (torch.float32, np.dtype('>f4'), 'f4'),
            14: (torch.float64, np.dtype('>f8'), 'f8')}
    # read
    with open_maybe_compressed_file(path) as f:
        data = f.read()
    # parse
    magic = get_int(data[0:4])
    nd = magic % 256
    ty = magic // 256
    assert nd >= 1 and nd <= 3
    assert ty >= 8 and ty <= 14
    m = read_sn3_pascalvincent_tensor.typemap[ty]
    s = [get_int(data[4 * (i + 1): 4 * (i + 2)]) for i in range(nd)]
    parsed = np.frombuffer(data, dtype=m[1], offset=(4 * (nd + 1)))
    assert parsed.shape[0] == np.prod(s) or not strict
    return torch.from_numpy(parsed.astype(m[2], copy=False)).view(*s)


def read_label_file(path):
    with open(path, 'rb') as f:
        x = read_sn3_pascalvincent_tensor(f, strict=False)
    assert(x.dtype == torch.uint8)
    assert(x.ndimension() == 1)
    return x.long()


def read_image_file(path):
    with open(path, 'rb') as f:
        x = read_sn3_pascalvincent_tensor(f, strict=False)
    assert(x.dtype == torch.uint8)
    assert(x.ndimension() == 3)
    return x



import torch
import torch.nn as nn
import torch.nn as nn
from torch.nn.parameter import Parameter
import numpy as np
import sys
from .utils import *

################ zhirui on 12-30-2020 ################
# this block is for buding the matrix of variant circuit
######################################################
import math


class VClassicCircuitMatrix:
    def __init__(self, n_qubits):
        # --- parameter definition ---
        self.n_qubits = n_qubits
        # state = state
        # constant matrix
        self.mat_cz = torch.tensor([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, -1]])
        self.mat_identity = torch.tensor([[1, 0], [0, 1]])
        self.mat_swap = torch.tensor([[1, 0, 0, 0], [0, 0, 1, 0], [0, 1, 0, 0], [0, 0, 0, 1]])

    def get_ry_matrix(self, theta):
        mm = torch.zeros(2, 2, dtype=torch.cdouble)
        mm = set_value(mm, 0, 0, torch.cos(theta / 2))
        mm = set_value(mm, 0, 1, -torch.sin(theta / 2))
        mm = set_value(mm, 1, 0, torch.sin(theta / 2))
        mm = set_value(mm, 1, 1, torch.cos(theta / 2))
        return mm

    def get_rx_matrix(self, theta):
        mm = torch.zeros(2, 2, dtype=torch.cdouble)
        mm = set_value(mm, 0, 0, torch.cos(theta / 2))
        mm = set_value(mm, 0, 1, -1j * torch.sin(theta / 2))
        mm = set_value(mm, 1, 0, -1j * torch.sin(theta / 2))
        mm = set_value(mm, 1, 1, torch.cos(theta / 2))
        return mm

    def get_rz_matrix(self, theta):
        mm = torch.zeros(2, 2, dtype=torch.cdouble)
        mm = set_value(mm, 0, 0, torch.exp(-1j * theta / 2))
        mm = set_value(mm, 1, 1, torch.exp(1j * theta / 2))
        return mm

    def get_ctrl_rx_matrix(self, theta):
        mm = torch.zeros(4, 4, dtype=torch.cdouble)
        mm = set_value(mm, 0, 0, torch.tensor(1, dtype=torch.cdouble))
        mm = set_value(mm, 1, 1, torch.tensor(1, dtype=torch.cdouble))
        mm = set_value(mm, 2, 2, torch.cos(theta / 2))
        mm = set_value(mm, 2, 3, -1j * torch.sin(theta / 2))
        mm = set_value(mm, 3, 2, -1j * torch.sin(theta / 2))
        mm = set_value(mm, 3, 3, torch.cos(theta / 2))
        return mm

    def get_ctrl_ry_matrix(self, theta):
        mm = torch.zeros(4, 4, dtype=torch.cdouble)
        mm = set_value(mm, 0, 0, torch.tensor(1, dtype=torch.cdouble))
        mm = set_value(mm, 1, 1, torch.tensor(1, dtype=torch.cdouble))
        mm = set_value(mm, 2, 2, torch.cos(theta / 2))
        mm = set_value(mm, 2, 3, -torch.sin(theta / 2))
        mm = set_value(mm, 3, 2, torch.sin(theta / 2))
        mm = set_value(mm, 3, 3, torch.cos(theta / 2))
        return mm

    def get_ctrl_rz_matrix(self, theta):
        mm = torch.zeros(4, 4, dtype=torch.cdouble)
        mm = set_value(mm, 0, 0, torch.tensor(1, dtype=torch.cdouble))
        mm = set_value(mm, 1, 1, torch.tensor(1, dtype=torch.cdouble))
        mm = set_value(mm, 2, 2, torch.exp(-1j * theta / 2))
        mm = set_value(mm, 3, 3, torch.exp(1j * theta / 2))
        return mm

    # swap (index) and (index-1)
    def qf_swap_dec(self, state, index):
        # generate the matrix
        temp_mat = torch.ones(1, dtype=torch.cdouble)
        for i in range(0, index - 1):
            temp_mat = torch.kron(self.mat_identity, temp_mat)
        temp_mat = torch.kron(self.mat_swap, temp_mat)
        for i in range(0, self.n_qubits - 1 - index):
            temp_mat = torch.kron(self.mat_identity, temp_mat)
        # change state
        state = torch.mm(temp_mat, state)
        return state

    def qf_ry(self, state, theta, index):

        temp_mat = torch.ones(1, dtype=torch.cdouble)
        if isinstance(index, int):
            for i in range(0, self.n_qubits):
                if i == index:
                    mm = self.get_ry_matrix(theta)
                    temp_mat = torch.kron(mm, temp_mat)
                else:
                    temp_mat = torch.kron(self.mat_identity, temp_mat)
        else:
            for i in range(0, self.n_qubits):
                if i in index:
                    select_theta = torch.index_select(theta, 0, torch.tensor([i]))
                    select_mm = self.get_ry_matrix(select_theta)
                    temp_mat = torch.kron(select_mm, temp_mat)

                else:
                    temp_mat = torch.kron(self.mat_identity, temp_mat)
                    # change state
        state = torch.mm(temp_mat, state)
        return state

    def qf_rx(self, state, theta, index):
        temp_mat = torch.ones(1, dtype=torch.cdouble)
        if isinstance(index, int):
            for i in range(0, self.n_qubits):
                if i == index:
                    mm = self.get_rx_matrix(theta)
                    temp_mat = torch.kron(mm, temp_mat)
                else:
                    temp_mat = torch.kron(self.mat_identity, temp_mat)
        else:
            for i in range(0, self.n_qubits):
                if i in index:
                    select_theta = torch.index_select(theta, 0, torch.tensor([i]))
                    select_mm = self.get_rx_matrix(select_theta)
                    temp_mat = torch.kron(select_mm, temp_mat)

                else:
                    temp_mat = torch.kron(self.mat_identity, temp_mat)
        state = torch.mm(temp_mat, state)
        return state

    def qf_rz(self, state, theta, index):
        temp_mat = torch.ones(1, dtype=torch.cdouble)
        if isinstance(index, int):
            for i in range(0, self.n_qubits):
                if i == index:
                    mm = self.get_rz_matrix(theta)
                    temp_mat = torch.kron(mm, temp_mat)
                else:
                    temp_mat = torch.kron(self.mat_identity, temp_mat)
        else:
            for i in range(0, self.n_qubits):
                if i in index:
                    select_theta = torch.index_select(theta, 0, torch.tensor([i]))
                    select_mm = self.get_rz_matrix(select_theta)
                    temp_mat = torch.kron(select_mm, temp_mat)

                else:
                    temp_mat = torch.kron(self.mat_identity, temp_mat)
        state = torch.mm(temp_mat, state)
        return state

    def qf_cz(self, state, index1, index2):

        # generate the matrix

        # swap the bottom one next to the up one
        for i in range(index2, index1 + 1, -1):
            state = self.qf_swap_dec(state, i)

        # generate cz matrix
        temp_mat = torch.ones(1, dtype=torch.cdouble)
        for i in range(0, index1):
            temp_mat = torch.kron(self.mat_identity, temp_mat)

        temp_mat = torch.kron(self.mat_cz, temp_mat)

        for i in range(0, self.n_qubits - 2 - index1):
            temp_mat = torch.kron(self.mat_identity, temp_mat)

        # change state
        state = torch.mm(temp_mat, state)

        # swap back
        for i in range(index1 + 2, index2 + 1):
            state = self.qf_swap_dec(state, i)

        return state

    def qf_c_r(self, state, theta, name, index_c, index_r):

        if name == 'x':
            self.mat_r = self.get_ctrl_rx_matrix(theta)
        elif name == 'y':
            self.mat_r = self.get_ctrl_ry_matrix(theta)
        elif name == 'z':
            self.mat_r = self.get_ctrl_rz_matrix(theta)

        if index_c > index_r:
            # swap the bottom one next to the up one
            for i in range(index_r + 1, index_c + 1):
                state = self.qf_swap_dec(state, i)

            # generate cz matrix
            temp_mat = torch.ones(1, dtype=torch.cdouble)

            for i in range(0, index_c - 1):
                temp_mat = torch.kron(self.mat_identity, temp_mat)

            temp_mat = torch.kron(self.mat_r, temp_mat)

            for i in range(0, self.n_qubits - 1 - index_c):
                temp_mat = torch.kron(self.mat_identity, temp_mat)

            # change state
            state = torch.mm(temp_mat, state)

            # swap back
            for i in range(index_c, index_r, -1):
                state = self.qf_swap_dec(state, i)
        else:
            # swap the bottom one next to the up one
            for i in range(index_r, index_c + 1, -1):
                state = self.qf_swap_dec(state, i)

            # generate cz matrix
            temp_mat = torch.ones(1, dtype=torch.cdouble)
            for i in range(0, index_c):
                temp_mat = torch.kron(self.mat_identity, temp_mat)

            temp_mat = torch.kron(self.mat_r, temp_mat)

            for i in range(0, self.n_qubits - 2 - index_c):
                temp_mat = torch.kron(self.mat_identity, temp_mat)

            # change state
            state = torch.mm(temp_mat, state)

            # swap back
            for i in range(index_c + 2, index_r + 1):
                state = self.qf_swap_dec(state, i)
        return state

    # the code is similar to the VQuantumCircuit::vqc_10
    def vqc_10(self, state, thetas):

        # head ry part
        state = self.qf_ry(state, thetas[0:self.n_qubits], range(0, self.n_qubits))

        # cz part
        for i in range(self.n_qubits - 1):
            state = self.qf_cz(state, self.n_qubits - 2 - i, self.n_qubits - 1 - i)
        state = self.qf_cz(state, 0, self.n_qubits - 1)

        # tail ry part
        state = self.qf_ry(state, thetas[self.n_qubits:2 * self.n_qubits], range(0, self.n_qubits))
        return state

    def vqc_5(self, state, thetas):
        state = self.qf_rx(state, thetas[0:self.n_qubits], range(0, self.n_qubits))
        state = self.qf_rz(state, thetas[self.n_qubits:2 * self.n_qubits], range(0, self.n_qubits))
        cnt = 0
        for i in range(self.n_qubits - 1, -1, -1):
            for j in range(self.n_qubits - 1, -1, -1):
                if j == i:
                    continue
                else:
                    state = self.qf_c_r(state, thetas[2 * self.n_qubits + cnt], 'z', i, j)
                    cnt = cnt + 1
        state = self.qf_rx(state, thetas[5 * self.n_qubits:6 * self.n_qubits], range(0, self.n_qubits))
        state = self.qf_rz(state, thetas[6 * self.n_qubits:7 * self.n_qubits], range(0, self.n_qubits))
        return state

    def get_parameter_number(self, vqc_name):
        if vqc_name == 'vqc_10':
            return int(2 * self.n_qubits)
        elif vqc_name == 'vqc_5':
            return int(7 * self.n_qubits)


class VQC_Net(nn.Module):
    def __init__(self, input_num, output_num, vqc_name='vqc_10'):
        super(VQC_Net, self).__init__()

        # init parameter
        self.num_qubit = int(math.log2(input_num))
        self.vcm = VClassicCircuitMatrix(self.num_qubit)
        self.output_num = output_num
        self.vqc_name = vqc_name
        self.theta = Parameter(
            torch.tensor(np.random.randn(self.vcm.get_parameter_number(vqc_name)) * np.pi, dtype=torch.cdouble,
                         requires_grad=True))  # [np.pi/3,np.pi/4,np.pi/3,np.pi/9,np.pi,np.pi/4,np.pi/10,np.pi/2]

        # init  VClassicCircuitMatrix

    def forward(self, x, training=True):

        x = x.t().to(torch.cdouble)
        if x.shape[0] != int(math.pow(2, self.num_qubit)):
            print("Not support VQC input size :", x.shape,"n_qubit :",self.num_qubit)
            sys.exit(0)

        if self.output_num > int(math.pow(2, self.num_qubit)):
            print("Not support VQC output size!")
            sys.exit(0)

        if self.vqc_name == 'vqc_10':
            x = self.vcm.vqc_10(x, self.theta)
        elif self.vqc_name == 'vqc_5':
            x = self.vcm.vqc_5(x, self.theta)
        else:
            print("Not support VQC name!")
            sys.exit(0)

        if self.output_num <= self.num_qubit:
            x = amp2prop(x)

        x = torch.index_select(x, 0, torch.tensor(range(self.output_num)))

        return x.t().float()



from .base import *


class N_LYR_Circ(BaseCircuit):
    """
    N_LYR_Circ is a class, which includes functions to build N-layer circuit(for batch normalization)

    Args:
         n_qubits: input qubits of each unit
         n_repeats: repeat times of each unit

    """
    def __init__(self, n_qubits, n_repeats=1):

        self.n_qubits = n_qubits
        self.n_repeats = n_repeats

    def add_norm_qubits(self, circuit):
        norm_qubits = self.add_qubits(circuit, "norm_qbits", self.n_qubits)
        return norm_qubits

    def add_out_qubits(self, circuit):
        out_qubits = self.add_qubits(circuit, "norm_output_qbits", self.n_qubits)
        return out_qubits

    def forward(self, circuit, input_qubits, norm_qubits, out_qubits, norm_flag, norm_para):
        """
        Function forward is to add the circuit of N-layer.

        Args:
             circuit: The  circuit that you add the unit at the end
             input_qubits: The register of input qubits
             norm_qubits: The register of qubits that do angle-encoding.
             out_qubits: The register of output qubits
             norm_flag: the direction of batch normalization.
             norm_para: the parameter of batch normalization  encoded in the angle of qubits.

        """
        for i in range(self.n_qubits):
            norm_init_rad = float(norm_para[i].sqrt().arcsin() * 2)
            circuit.ry(norm_init_rad, norm_qubits[i])
            if norm_flag[i]:
                circuit.cx(input_qubits[i], out_qubits[i])
                circuit.x(input_qubits[i])
                circuit.ccx(input_qubits[i], norm_qubits[i], out_qubits[i])
            else:
                circuit.ccx(input_qubits[i], norm_qubits[i], out_qubits[i])



from .utils import *




class indiv_adj(nn.Module):
    def __init__(self, num_features, init_ang_inc=1, momentum=0.1,training = False):
        super(indiv_adj, self).__init__()

        self.x_running_rot = Parameter(torch.zeros((num_features)), requires_grad=False)
        # self.ang_inc = Parameter(torch.ones(1)*init_ang_inc)
        self.ang_inc = Parameter(torch.tensor(init_ang_inc,dtype=torch.float32),requires_grad=True)
        self.momentum = momentum

        self.printed = False
        self.x_mean_ancle = 0
        self.x_mean_rote = 0
        self.input = 0
        self.output = 0

    def forward(self, x, training=True):
        if not training:
            if not self.printed:
                # print("self.ang_inc", self.ang_inc)
                self.printed = True
            x_1 = (self.x_running_rot * x)

        else:
            self.printed = False
            x = x.transpose(0, 1)
            x_sum = x.sum(-1).unsqueeze(-1).expand(x.shape)
            x_lack_sum = x_sum + x
            x_mean = x_lack_sum / x.shape[-1]

            ang_inc = (self.ang_inc > 0).float() * self.ang_inc + 1

            y = 0.5 / x_mean
            y = y.transpose(0, 1)
            y = y / ang_inc
            y = y.transpose(0, 1)

            x_moving_rot = (y.sum(-1) / x.shape[-1])

            self.x_running_rot[:] = self.momentum * self.x_running_rot + \
                                    (1 - self.momentum) * x_moving_rot

            x_1 = y * x
            x_1 = x_1.transpose(0, 1)

        return x_1

    def reset_parameters(self):
        self.reset_running_stats()
        self.ang_inc.data.zeros_()


class batch_adj(nn.Module):
    def __init__(self, num_features, momentum=0.1):
        super(batch_adj, self).__init__()
        self.x_running_rot = Parameter(torch.zeros(num_features), requires_grad=False)
        self.momentum = momentum
        self.x_l_0_5 = Parameter(torch.zeros(num_features), requires_grad=False)
        self.x_g_0_5 = Parameter(torch.zeros(num_features), requires_grad=False)

    def forward(self, x, training=True):
        if not training:
            x_1 = self.x_l_0_5 * (self.x_running_rot * (1 - x) + x)
            x_1 += self.x_g_0_5 * (self.x_running_rot * x)
        else:
            x = x.transpose(0, 1)
            x_sum = x.sum(-1)
            x_mean = x_sum / x.shape[-1]

            self.x_l_0_5[:] = ((x_mean <= 0.5).float())
            self.x_g_0_5[:] = ((x_mean > 0.5).float())

            y = self.x_l_0_5 * ((0.5 - x_mean) / (1 - x_mean))
            y += self.x_g_0_5 * (0.5 / x_mean)

            self.x_running_rot[:] = self.momentum * self.x_running_rot + \
                                    (1 - self.momentum) * y

            x = x.transpose(0, 1)
            x_1 = self.x_l_0_5 * (y * (1 - x) + x)
            x_1 += self.x_g_0_5 * (y * x)

        return x_1

class N_LYR(nn.Module):
    def __init__(self, num_features, init_ang_inc=1, momentum=0.1,training = False):
        super(N_LYR, self).__init__()
        self.inv = indiv_adj(num_features, init_ang_inc, momentum,training)
        self.batch = batch_adj(num_features,momentum)
    def forward(self, x, training=True):
        x = self.batch(x,training)
        x = self.inv(x,training)
        return x



import sys
import numpy as np
import numpy as np
from qiskit.tools.monitor import job_monitor
from qiskit import QuantumRegister
from qiskit.extensions import  UnitaryGate
from qiskit import Aer, execute,IBMQ,transpile
import math
from qiskit import BasicAer
import copy

from .gates import *
from .base import *

class V_LYR_Circ(BaseCircuit):
    """
    V_LYR_Circ is a class, which includes functions to build v-layer(vqc)

    Args:
         n_qubits: input qubits of each unit
         n_repeats: repeat times of each unit

    """
    def __init__(self, n_qubits,n_repeats):
        self.n_qubits = n_qubits
        self.n_repeats = n_repeats



    def add_input_qubits(self,circuit,name="vqc_input"):
        inps = BaseCircuit.add_input_qubits(self,circuit,name)
        return inps



    #define the circuit
    def vqc_10(self,circuit,input_qubits,thetas):
        # print(input_qubits)
        #head ry part 
        for i in range(0,self.n_qubits):
            circuit.ry(thetas[i], input_qubits[i])
        circuit.barrier(input_qubits)
        
        #cz part
        for i in range(self.n_qubits-1):
            circuit.cz(input_qubits[self.n_qubits-2-i],input_qubits[self.n_qubits-1-i])
        circuit.cz(input_qubits[0],input_qubits[self.n_qubits-1])
        circuit.barrier(input_qubits)

        #tail ry part
        for i in range(0,self.n_qubits):
            circuit.ry(thetas[i+self.n_qubits], input_qubits[i])


    def vqc_5(self,circuit,input_qubits,thetas):
        for i in range(0,self.n_qubits):
            circuit.rx(thetas[i],input_qubits[i])
        for i in range(0,self.n_qubits):
            circuit.rz(thetas[self.n_qubits+i],input_qubits[i])
        
        circuit.barrier(input_qubits)
        cnt = 0
        for i in range(self.n_qubits-1,-1,-1):
            for j in range(self.n_qubits-1,-1,-1):
                if j == i:
                    continue
                else:
                    circuit.crz(thetas[2*self.n_qubits + cnt],input_qubits[i],input_qubits[j])
                    cnt = cnt +1
            circuit.barrier(input_qubits)
        for i in range(0,self.n_qubits):
            circuit.rx(thetas[5*self.n_qubits+i],input_qubits[i])
        for i in range(0,self.n_qubits):
            circuit.rz(thetas[6*self.n_qubits+i],input_qubits[i])

    def get_parameter_number(self,vqc_name):
        if vqc_name == 'v10':
            return int(2*self.n_qubits)
        elif vqc_name == 'v5':
            return int(7*self.n_qubits)

    
    def forward(self,circuit,input_qubits,vqc_name,thetas):
        """
        Function forward is to add the circuit of VQC.

        Args:
             circuit: The  circuit that you add the unit at the end
             input_qubits: The register of input qubits
             vqc_name: The number of vqc architecture in the  paper.
             thetas: A list of angles to be encoded into  of vqc architecture in the paper.
             aux: aux qubits
        """

        if vqc_name == 'v10':
            for i in range(self.n_repeats):
                self.vqc_10(circuit,input_qubits[i],thetas)
        elif vqc_name == 'v5':
            for i in range(self.n_repeats):
                self.vqc_5(circuit,input_qubits[i],thetas)



import math

'''
This is the implementation of Algorithm 4 in Box2 of QuantumFlow paper at NCOMM (Weiwen)
'''

def print_info():
    print("This is in qf_map")

def change_sign(sign, bin):
    one_positions = []
    try:
        beg_pos = 0
        while True:
            find_pos = bin.index("1", beg_pos)
            one_positions.append(find_pos)
            beg_pos = find_pos + 1
    except Exception as exception:
        # print("Not Found")
        pass
    for k, v in sign.items():
        change = True
        for pos in one_positions:
            if k[pos] == "0":
                change = False
                break
        if change:
            sign[k] = -1 * v


def find_start(affect_count_table, target_num):
    for k in list(affect_count_table.keys())[::-1]:
        if target_num <= k:
            return k

def recursive_change(direction, start_point, target_num, sign, affect_count_table, quantum_gates):
    if start_point == target_num:
        # print("recursive_change: STOP")
        return

    gap = int(math.fabs(start_point - target_num))
    step = find_start(affect_count_table, gap)
    change_sign(sign, affect_count_table[step])
    quantum_gates.append(affect_count_table[step])

    if direction == "r":
        # print("recursive_change: From",start_point,"Right(-):",step)
        start_point = start_point - step
        direction = "l"
        recursive_change(direction, start_point, target_num, sign, affect_count_table, quantum_gates)

    else:
        # print("recursive_change: From",start_point,"Left(+):",step)
        start_point = start_point + step
        direction = "r"
        recursive_change(direction, start_point, target_num, sign, affect_count_table, quantum_gates)

def Mapping_U_LYR(sign, target_num, digits):
    flag = "0" + str(digits) + "b"
    pre_num = 0
    affect_count_table = {}
    quantum_gates = []
    for i in range(digits):
        cur_num = pre_num + pow(2, i)
        pre_num = cur_num
        binstr_cur_num = format(cur_num, flag)
        affect_count_table[int(pow(2, binstr_cur_num.count("0")))] = binstr_cur_num

    if target_num in affect_count_table.keys():
        quantum_gates.append(affect_count_table[target_num])
        change_sign(sign, affect_count_table[target_num])

    else:
        direction = "r"
        start_point = find_start(affect_count_table, target_num)
        quantum_gates.append(affect_count_table[start_point])
        change_sign(sign, affect_count_table[start_point])
        recursive_change(direction, start_point, target_num, sign, affect_count_table, quantum_gates)

    return quantum_gates



from .utils import *

class B_LYR(nn.Linear):

    def forward(self, input, training=True):
        binary_weight = binarize(self.weight)
        if self.bias is None:
            output = F.linear(input, binary_weight)
            # print(input,binary_weight, input.shape[-1])
            output = torch.div(output, input.shape[-1])
            # output = torch.pow(output, 2)

            return output
        else:
            print("Not Implement")
            sys.exit(0)

    def reset_parameters(self):
        # Glorot initialization
        in_features, out_features = self.weight.size()
        stdv = math.sqrt(1.5 / (in_features + out_features))
        self.weight.data.uniform_(-stdv, stdv)
        if self.bias is not None:
            self.bias.data.zero_()

        self.weight.lr_scale = 1. / stdv




import torch
import torchvision.transforms as transforms
from torchvision import datasets
import numpy as np
from .c_mnist import *


def modify_target_ori(target, interest_num):
    for j in range(len(target)):
        for idx in range(len(interest_num)):
            if target[j] == interest_num[idx]:
                target[j] = idx
                break

    new_target = torch.zeros(target.shape[0], len(interest_num))

    for i in range(target.shape[0]):
        one_shot = torch.zeros(len(interest_num))
        one_shot[target[i].item()] = 1
        new_target[i] = one_shot.clone()

    return target, new_target


def modify_target(target, interest_num):
    new_target = torch.zeros(target.shape[0], len(interest_num))

    for i in range(target.shape[0]):
        one_shot = torch.zeros(len(interest_num))
        one_shot[target[i].item()] = 1
        new_target[i] = one_shot.clone()
    return target, new_target


def select_num(dataset, interest_num):
    """
    Function select_num is to select the specfic label of a dataset to generate a sub-dataset.
    Args:
         dataset: a pytorch-datasets
         interest_num: the specfic label list,such as [3,6]
    Returns:
        dataset: a sub-dataset
    """
    labels = dataset.targets  # get labels
    labels = labels.numpy()
    idx = {}
    for num in interest_num:
        idx[num] = np.where(labels == num)

    fin_idx = idx[interest_num[0]]
    for i in range(1, len(interest_num)):
        fin_idx = (np.concatenate((fin_idx[0], idx[interest_num[i]][0])),)

    fin_idx = fin_idx[0]

    dataset.targets = labels[fin_idx]
    dataset.data = dataset.data[fin_idx]

    # print(dataset.targets.shape)

    dataset.targets, _ = modify_target_ori(dataset.targets, interest_num)
    # print(dataset.targets.shape)

    return dataset


class ToQuantumData_Batch(object):

    def __call__(self, tensor):
        data = tensor
        input_vec = data.view(-1)
        vec_len = input_vec.size()[0]
        input_matrix = torch.zeros(vec_len, vec_len)
        input_matrix[0] = input_vec
        input_matrix = input_matrix.transpose(0, 1)
        u, s, v = np.linalg.svd(input_matrix)
        output_matrix = torch.tensor(np.dot(u, v))
        output_data = output_matrix[:, 0].view(data.shape)
        return output_data


class ToQuantumData(object):
    """
    class ToQuantumData is to transform the input image into a unitary matrix to be encoded in quantum unitary matrix.
    """
    def __init__(self, img_size):
        super().__init__()
        self.img_size = img_size

    def __call__(self, tensor):
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        data = tensor.to(device)
        input_vec = data.view(-1)
        vec_len = input_vec.size()[0]
        input_matrix = torch.zeros(vec_len, vec_len)
        input_matrix[0] = input_vec
        input_matrix = input_matrix.transpose(0, 1)
        u, s, v = np.linalg.svd(input_matrix)
        output_matrix = torch.tensor(np.dot(u, v))
        # print(output_matrix)
        output_data = output_matrix[:, 0].view(1, self.img_size, self.img_size)
        return output_data


def load_data(interest_num, datapath, isppd, img_size, batch_size, inference_batch_size, is_to_q=True, num_workers=0):
    """
    Function load_data is to get the DataLoader of MNIST after preprocessing with the interest nums.

    Args:
         interest_num: the specfic label list,such as [3,6]
         datapath: the path of MNIST data
         isppd: whether the data is loaded from qfMNIST,whch have done preprocessing
         img_size: the width /height of input image (width = height)
         batch_size:  batch size while training
         inference_batch_size: batch size while inferencing
         is_to_q: whether the output data in loader is to be encoded in quantum unitary matrix
         num_workers: for torch.utils.data.DataLoader.

    Returns:
        train_loader: a torch.utils.data.DataLoader for training.
        test_loader: a torch.utils.data.DataLoader for testing.
    """
    if isppd:
        train_data = qfMNIST(root=datapath, img_size=img_size, train=True)
        test_data = qfMNIST(root=datapath, img_size=img_size, train=False)

    else:
        # convert data to torch.FloatTensor
        if is_to_q:
            transform = transforms.Compose(
                [transforms.Resize((img_size, img_size)), transforms.ToTensor(), ToQuantumData(img_size)])
            transform_inference = transforms.Compose(
                [transforms.Resize((img_size, img_size)), transforms.ToTensor(), ToQuantumData(img_size)])
        else:
            transform = transforms.Compose([transforms.Resize((img_size, img_size)), transforms.ToTensor()])
            transform_inference = transforms.Compose([transforms.Resize((img_size, img_size)), transforms.ToTensor()])

        # choose the training and test datasets
        train_data = datasets.MNIST(root=datapath, train=True,
                                    download=True, transform=transform)
        test_data = datasets.MNIST(root=datapath, train=False,
                                   download=True, transform=transform_inference)

    train_data = select_num(train_data, interest_num)
    test_data = select_num(test_data, interest_num)

    # prepare data loaders
    train_loader = torch.utils.data.DataLoader(train_data, batch_size=batch_size,
                                               num_workers=num_workers, shuffle=True, drop_last=True)
    test_loader = torch.utils.data.DataLoader(test_data, batch_size=inference_batch_size,
                                              num_workers=num_workers, shuffle=True, drop_last=True)

    return train_loader, test_loader


def to_quantum_matrix(tensor):
    """
    Function to_quantum_matrix is to transform the input image into a unitary matrix.
    Args:
         tensor: input image
    Returns:
        output_matrix: a unitary matrix.
    """
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    data = tensor.to(device)
    input_vec = data.view(-1)
    vec_len = input_vec.size()[0]
    input_matrix = torch.zeros(vec_len, vec_len)
    input_matrix[0] = input_vec
    input_matrix = np.float64(input_matrix.transpose(0,1))
    u, s, v = np.linalg.svd(input_matrix)
    output_matrix = torch.tensor(np.dot(u, v), dtype=torch.float64)
    return output_matrix 


import sys

class ExtendGate():

    ################ Weiwen on 06-02-2021 ################
    # Function: ccz from Listing 3
    # Note: using the basic Toffoli gates and CZ gate
    #       to implement ccz gate, which will flip the
    #       sign of state |111>
    # Parameters: (1) quantum circuit;
    #             (2-3) control qubits;
    #             (4) target qubits;
    #             (5) auxiliary qubits.
    ######################################################
    @classmethod
    def ccz(cls, circ, q1, q2, q3, aux1):
        """
        Function ccz : using the basic Toffoli gates and CZ gate to implement ccz gate, which will flip the sign of state |111>

        Args:
             circ: quantum circuit
             q1,q2: control qubits
             q3: target qubits :
             aux1: auxiliary qubits
        Returns:
             circ:  the circuit you add the gate to

        """

        # Apply Z-gate to a state controlled by 3 qubits
        circ.ccx(q1, q2, aux1)
        circ.cz(aux1, q3)
        # cleaning the aux bit
        circ.ccx(q1, q2, aux1)
        return circ

    @classmethod
    def cccx(cls, circ, q1, q2, q3, q4, aux1):
        """
        Function cccx : using the basic Toffoli gates and CX gate to implement cccx gate
        Args:
             circ: quantum circuit
             q1,q2,q3: control qubits
             q4: target qubits :
             aux1: auxiliary qubits
        Returns:
             circ:  the circuit you add the gate to

        """
        # Apply Z-gate to a state controlled by 3 qubits
        circ.ccx(q1, q2, aux1)
        circ.ccx(q3, aux1, q4)
        # cleaning the aux bits
        circ.ccx(q1, q2, aux1)
        return circ

    ################ Weiwen on 12-30-2020 ################
    # Function: cccz from Listing 3
    # Note: using the basic Toffoli gates and CZ gate
    #       to implement cccz gate, which will flip the
    #       sign of state |1111>
    # Parameters: (1) quantum circuit;
    #             (2-4) control qubits;
    #             (5) target qubits;
    #             (6-7) auxiliary qubits.
    ######################################################
    @classmethod
    def cccz(cls, circ, q1, q2, q3, q4, aux1, aux2):
        """
        Function cccx : using the basic Toffoli gates and CZ gate to implement cccz gate

        Args:
             circ: quantum circuit
             q1,q2,q3: control qubits
             q4: target qubits :
             aux1,aux2: auxiliary qubits

        Returns:
             circ:  the circuit you add the gate to

        """
        # Apply Z-gate to a state controlled by 4 qubits
        circ.ccx(q1, q2, aux1)
        circ.ccx(q3, aux1, aux2)
        circ.cz(aux2, q4)
        # cleaning the aux bits
        circ.ccx(q3, aux1, aux2)
        circ.ccx(q1, q2, aux1)
        return circ

    @classmethod
    def cnz(cls, circ, q, aux, q_num):
        """
        Function cnz : using the basic Toffoli gates and CZ gate to implement cc..cz gate,which includes n control gates.

        Args:
             circ: quantum circuit
             q:  qubits list. The last is target qubits
             aux: auxiliary qubits
             q_num: the number of qubits in q.
        """

        if q_num<=2:
            print("Please use cz instead of cnz!")
            sys.exit(0)
        else:
            ccx_list = []
            p0 = q[0]
            p1 = q[1]
            p2 = aux[0]
            circ.ccx(p0, p1, p2)
            ccx_list.append((p0,p1,p2))
            for i in range(2,q_num-1):
                p0 = q[i]
                p1 = p2
                p2 = aux[i-1]
                circ.ccx(p0, p1, p2)
                ccx_list.append((p0, p1, p2))
            circ.cz(aux[q_num-3],q[q_num-1])

            for gate in reversed(ccx_list):
                circ.ccx(gate[0],gate[1],gate[2])

    @classmethod
    def cnx(cls, circ, q, aux, q_num):
        """
        Function cnz : using the basic Toffoli gates and CZ gate to implement cc..cx gate,which includes n control gates.

        Args:
             circ: quantum circuit
             q:  qubits list. The last is target qubits
             aux: auxiliary qubits
             q_num: the number of qubits in q.
        """

        if q_num <= 3:
            print("Please use ccx instead of cnx!")
            sys.exit(0)
        else:
            ccx_list = []
            p0 = q[0]
            p1 = q[1]
            p2 = aux[0]
            circ.ccx(p0, p1, p2)
            ccx_list.append((p0, p1, p2))
            for i in range(2, q_num - 2):
                p0 = q[i]
                p1 = p2
                p2 = aux[i - 1]
                circ.ccx(p0, p1, p2)
                ccx_list.append((p0, p1, p2))
            circ.ccx(aux[q_num - 4], q[q_num - 2], q[q_num - 1])

            for gate in reversed(ccx_list):
                circ.ccx(gate[0], gate[1], gate[2])

    ################ Weiwen on 12-30-2020 ################
    # Function: cccz from Listing 4
    # Note: using the basic Toffoli gate to implement ccccx
    #       gate. It is used to switch the quantum states
    #       of |11110> and |11111>.
    # Parameters: (1) quantum circuit;
    #             (2-5) control qubits;
    #             (6) target qubits;
    #             (7-8) auxiliary qubits.
    ######################################################
    @classmethod
    def ccccx(cls, circ, q1, q2, q3, q4, q5, aux1, aux2):
        """
        Function cccx : using the basic Toffoli gates and CX,CCX gate to implement ccccx gate

        Args:
             circ: quantum circuit
             q1,q2,q3,q4: control qubits
             q5: target qubits :
             aux1,aux2: auxiliary qubits

        Returns:
             circ:  the circuit you add the gate to

        """
        circ.ccx(q1, q2, aux1)
        circ.ccx(q3, q4, aux2)
        circ.ccx(aux2, aux1, q5)
        # cleaning the aux bits
        circ.ccx(q3, q4, aux2)
        circ.ccx(q1, q2, aux1)
        return circ

    ################ Weiwen on 12-30-2020 ################
    # Function: neg_weight_gate from Listing 3
    # Note: adding NOT(X) gate before the qubits associated
    #       with 0 state. For example, if we want to flip
    #       the sign of |1101>, we add X gate for q2 before
    #       the cccz gate, as follows.
    #       --q3-----|---
    #       --q2----X|X--
    #       --q1-----|---
    #       --q0-----z---
    # Parameters: (1) quantum circuit;
    #             (2) all qubits, say q0-q3;
    #             (3) the auxiliary qubits used for cccz
    #             (4) states, say 1101
    ######################################################
    @classmethod
    def neg_weight_gate(cls, circ, qubits, aux, state):
        """
        Function neg_weight_gate : adding NOT(X) gate before the qubits associated with 0 state. For example, if we want to flip the sign of |1101>, we add X gate for q2 beforethe cccz gate, as follows.

        Args:
             circ: quantum circuit
             qubits: the register of input qubits
             aux: auxiliary qubits
             state: a  string that represents the state of the qubits,such as '1101'

        """
        idx = 0
        # The index of qubits are reversed in terms of states.
        # As shown in the above example: we put X at q2 not the third position.
        state = state[::-1]
        for idx in range(len(state)):
            if state[idx] == '0':
                circ.x(qubits[idx])
        if len(state) ==1:
            circ.z(qubits[0])
        elif len(state) ==2:
            circ.cz(qubits[0], qubits[1])
        else:
            cls.cnz(circ,qubits,aux,len(state))
        #cls.cccz(circ, qubits[0], qubits[1], qubits[2], qubits[3], aux[0], aux[1])
        for idx in range(len(state)):
            if state[idx] == '0':
                circ.x(qubits[idx])

# if __name__ == "__main__":
#     from qiskit import QuantumRegister, QuantumCircuit
#     import warnings

#     warnings.filterwarnings("ignore")
#     input = QuantumRegister(3, "in")
#     aux = QuantumRegister(1, "aux")
#     circ = QuantumCircuit(input, aux)
#     # ExtendGate.cnx(circ,input,aux,3)
#     # circ.barrier()
#     ExtendGate.cnz(circ, input, aux, 3)
#     print(circ)


from math import log2,sqrt,asin
from .base import *
from .gates import ExtendGate 
# just for temp
class P_LYR_Circ(LinnerCircuit):
    """
    P_LYR_Circ is a class, which includes functions to build p-layer circuit

    Args:
         n_qubits: input qubits of each unit
         n_repeats: repeat times of each unit

    """
    def __init__(self, n_qubits, n_repeats):

        self.n_qubits = n_qubits
        self.n_repeats = n_repeats
        self.n_log = log2(self.n_qubits)

        if self.n_qubits == 1 :
            print('PLayerCircuit: The input size does not support 1 !')
            sys.exit(0)
        if abs(self.n_log - round(self.n_log))>1e-2:
            print('PLayerCircuit: The input size must be 2^n!')


    def add_out_qubits(self, circuit,name = "p_layer_qbits"):
        out_qubits = self.add_qubits(circuit, name, self.n_repeats)
        return out_qubits

    def add_aux(self, circuit,name ="aux_qbit"):
        """
        Function add  add_aux is to add a group of qubits as input qubit .

        Args:
             circuit: The  circuit that you add the unit at the end

        Returns:
             qubits: The register of qubits

        """
        if self.n_qubits < 4:
            return None
        # TODO: 09/30, Potential bug.
        else:
            # q_num = round(self.n_log) + 1
            # aux_num = q_num -1
            aux = self.add_qubits(circuit,name , round(self.n_log)*(1+self.n_repeats))
        # else:
        #     print('The input size is too big. Qubits should be less than 4.')
        #     sys.exit(0)
        return aux


    def forward(self, circuit, weight, in_qubits, out_qubits,aux = []):
        """
        Function add forward is to add the circuit of batch normalization.

        Args:
             circuit: The  circuit that you add the unit at the end
             weight: A list of binary weight.
             in_qubits: The register of input qubits
             out_qubit: The register of output qubits
             aux: aux qubits

        """
        for i in range(self.n_repeats):
            # mul weight
            if weight[i].sum() < 0:
                weight[i] = weight[i] * -1
            idx = 0
            for idx in range(weight[i].flatten().size()[0]):
                if weight[i][idx] == -1:
                    circuit.x(in_qubits[idx])
            circuit.barrier()

            # sum and pow2
            if self.n_qubits == 2:
                circuit.h(out_qubits[i])
                circuit.x(out_qubits[i])
                circuit.cz(in_qubits[0], out_qubits[i])
                circuit.cz(in_qubits[1], out_qubits[i])
                circuit.h(out_qubits[i])
                circuit.x(out_qubits[i])
            else:
                #add weight
                operate_qubits = []
                aux_qubits = []
                encoding_num = round(self.n_log)
                operate_qubits.append(in_qubits[self.n_qubits-1])
                for k in range(encoding_num):
                    operate_qubits.append(aux[k+encoding_num*i])
                for k in range(encoding_num):
                    aux_qubits.append(aux[encoding_num+encoding_num*i+k])
                circuit.h(operate_qubits[1:])
                for j in range(self.n_qubits):
                    operate_qubits[0]=in_qubits[self.n_qubits-1-j]
                    state = "{0:b}".format(self.n_qubits+j).zfill(encoding_num)
                    state = state[::-1]
                    ExtendGate.neg_weight_gate(circuit,operate_qubits,aux_qubits,state)
                    circuit.barrier()

                #sum2
                circuit.h(operate_qubits[1:])
                circuit.x(operate_qubits[1:])


                qbits = operate_qubits[1:]
                if encoding_num == 2:
                    circuit.ccx(qbits[0], qbits[1], out_qubits[i])
                else:
                    operate_qubits2 = []
                    aux_qubits2 = []
                    for k in range(encoding_num):
                        operate_qubits2.append(qbits[k])
                        if k < encoding_num - 2:
                            aux_qubits2.append(aux_qubits[k])
                    operate_qubits2.append(out_qubits[i])
                    ExtendGate.cnx(circuit, operate_qubits2, aux_qubits2, encoding_num+1)

            # recover
            for idx in range(weight[i].flatten().size()[0]):
                if weight[i][idx] == -1:
                    circuit.x(in_qubits[idx])
            circuit.barrier()



class P_Neuron_Circ(P_LYR_Circ):
    """
    P_Neuron_Circ is a class, which includes functions to build P-Neuron circuit. P-Layer consists of serverl P-Neuron.A neuron only repeats onece.

    Args:
         n_qubits: input qubits of each unit

    """
    def __init__(self, n_qubits):
        self.n_qubits = n_qubits
        self.n_repeats = 1

        P_LYR_Circ.__init__(self,n_qubits,self.n_repeats)


    def forward(self, circuit, weight, in_qubits, out_qubits,aux = [],ang=[]):
        """
        Function forward is to add the circuit of batch normalization.

        Args:
             circuit: The  circuit that you add the unit at the end
             weight: A list of binary weight.
             in_qubits: The register of input qubits
             out_qubit: The register of output qubits
             ang: the angle list of angle encoding
             aux: aux qubits

        """
        for i in range (self.n_qubits):
            if len(ang)>i:
                angle = float(2*asin(sqrt(ang[i])))
                circuit.ry(angle, in_qubits[i])

        for i in range(self.n_repeats):
            # mul weight
            if weight[i].sum() < 0:
                weight[i] = weight[i] * -1
            idx = 0
            for idx in range(weight[i].flatten().size()[0]):
                if weight[i][idx] == -1:
                    circuit.x(in_qubits[idx])
            circuit.barrier()
            # TODO: Potential bug for P-LYR

            # sum and pow2
            if self.n_qubits == 2:
                circuit.h(out_qubits[i])
                circuit.x(out_qubits[i])
                circuit.cz(in_qubits[0], out_qubits[i])
                circuit.cz(in_qubits[1], out_qubits[i])
                circuit.h(out_qubits[i])
                circuit.x(out_qubits[i])
            else:
                #add weight
                operate_qubits = []
                aux_qubits = []
                encoding_num = round(self.n_log)
                operate_qubits.append(in_qubits[self.n_qubits-1])
                for k in range(encoding_num):
                    operate_qubits.append(aux[k])
                for k in range(encoding_num):
                    aux_qubits.append(aux[encoding_num+k])
                circuit.h(operate_qubits[1:])
                for j in range(self.n_qubits):
                    operate_qubits[0]=in_qubits[self.n_qubits-1-j]
                    state = "{0:b}".format(self.n_qubits+j).zfill(encoding_num)
                    state = state[::-1]
                    ExtendGate.neg_weight_gate(circuit,operate_qubits,aux_qubits,state)
                    circuit.barrier()

                #sum2
                circuit.h(operate_qubits[1:])
                circuit.x(operate_qubits[1:])


                qbits = operate_qubits[1:]
                if encoding_num == 2:
                    circuit.ccx(qbits[0], qbits[1], out_qubits[i])
                else:
                    operate_qubits2 = []
                    aux_qubits2 = []
                    for k in range(encoding_num):
                        operate_qubits2.append(qbits[k])
                        if k < encoding_num - 2:
                            aux_qubits2.append(aux_qubits[k])
                    operate_qubits2.append(out_qubits[i])
                    ExtendGate.cnx(circuit, operate_qubits2, aux_qubits2, encoding_num+1)
